/* main.c */
#include <tk.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

int Tk_AppInit(Tcl_Interp *interp);
int FilterObjCmd(ClientData clientData, Tcl_Interp *interp,
					int objc, Tcl_Obj *CONST objv[]);
int DetachObjCmd(ClientData clientData, Tcl_Interp *interp,
					int objc, Tcl_Obj *CONST objv[]);
int TraceObjCmd(ClientData clientData, Tcl_Interp *interp,
					int objc, Tcl_Obj *CONST objv[]);
int AttachObjCmd(ClientData clientData, Tcl_Interp *interp,
					int objc, Tcl_Obj *CONST objv[]);
int service_gui(void);

extern char *traceAttach(const char *fname, int pid);
extern void traceFile(int pid, FILE *fPtr);
extern void traceDetach(int pid);
/*****************************************************************************/

static Tcl_Interp *interp;
static int attached = 0;
static int quit = 1;
static int pid = 0;
const char *tmpf = "/tmp/ftrace.out";

int main(int argc, char *argv[]) {
	char *script[2];

	script[0] = argv[0];
	script[1] = "ftrace.tk";
	Tk_Main(2, script, Tk_AppInit);
	return 0;
}

int Tk_AppInit(Tcl_Interp *new_interp) {
   interp = new_interp;
   /*
    * Initialize packages
    */
   if (Tcl_Init(interp) == TCL_ERROR) {
      return TCL_ERROR;
   }
   if (Tk_Init(interp) == TCL_ERROR) {
      return TCL_ERROR;
   }
   /*
    * Define application-specific commands here.
    */
   Tcl_CreateObjCommand(interp, "trace_attach", AttachObjCmd,
      (ClientData)NULL, NULL);
   Tcl_CreateObjCommand(interp, "trace_run", TraceObjCmd,
      (ClientData)NULL, NULL);
   Tcl_CreateObjCommand(interp, "trace_detach", DetachObjCmd,
      (ClientData)NULL, NULL);
   Tcl_CreateObjCommand(interp, "trace_filt", FilterObjCmd,
      (ClientData)NULL, NULL);

   return TCL_OK;
}

char * getExecPath (char * path, size_t dest_len)
{
	ssize_t len;
	char filename[127];
	/* Build the filename
	 */
	snprintf(filename, sizeof(filename), "/proc/%d/exe", pid);
    len = readlink (filename, path, dest_len);
    if(len != -1)
	{
		path[len] = '\0';
        return path;
    }
	return NULL;
}

int FilterObjCmd(ClientData clientData, Tcl_Interp *inter,
					int objc, Tcl_Obj *CONST objv[])
{
	FILE *inPtr, *outPtr;
	char buf[255];
	char *sym  = Tcl_GetString(objv[1]);
	char *tmpfile = "/tmp/ftrace123";

	inPtr = fopen (tmpfile, "r");
	outPtr = fopen (tmpf, "w");
	while (fgets (buf, 255, inPtr) != NULL)
	{
		if (strstr (buf, sym) == NULL)
			fputs(buf, outPtr);
	}

	fclose(inPtr);
	fclose(outPtr);
	return TCL_OK;
}

int DetachObjCmd(ClientData clientData, Tcl_Interp *inter,
					int objc, Tcl_Obj *CONST objv[])
{
	quit = 1;
	if (!attached)
		return TCL_OK;
	attached = 0;
	traceDetach(pid);
	return TCL_OK;
}

int TraceObjCmd(ClientData clientData, Tcl_Interp *inter,
					int objc, Tcl_Obj *CONST objv[])
{
	Tcl_Obj *objPtr;
	FILE *fPtr;

	if (!attached || !quit)
	{
		char *err = "Not attached or already running!";
		objPtr = Tcl_NewStringObj(err, strlen(err));
		Tcl_SetObjResult(interp, objPtr);
		return TCL_ERROR;
	}

	fPtr = fopen (tmpf, "w");
	quit = 0;
	traceFile(pid, fPtr);
	quit = 1;
	fclose(fPtr);
	objPtr = Tcl_NewStringObj(tmpf, strlen(tmpf));
	Tcl_SetObjResult(interp, objPtr);
	return TCL_OK;
}

int AttachObjCmd(ClientData clientData, Tcl_Interp *inter,
					int objc, Tcl_Obj *CONST objv[])
{
	Tcl_Obj *objPtr;
	char fname[255];
	char *symStr;
	pid = atoi(Tcl_GetString(objv[1]));

	if (attached)
	{
		char *err = "Already attached!";
		objPtr = Tcl_NewStringObj(err, strlen(err));
		Tcl_SetObjResult(interp, objPtr);
		return TCL_ERROR;
	}
	if (getExecPath(fname, 255) != NULL)
	{
		if ((symStr = traceAttach(fname, pid)) != NULL)
		{
			attached = 1;
			objPtr = Tcl_NewStringObj(symStr, strlen(symStr));
			Tcl_SetObjResult(interp, objPtr);
		}
		return TCL_OK;
	}
	else
	{
		char *err = "Couldn't find pid";
		objPtr = Tcl_NewStringObj(err, strlen(err));
		Tcl_SetObjResult(interp, objPtr);
	}
	return TCL_ERROR;
}

int service_gui(void)
{
	if (atoi(Tcl_GetVar(interp, "tr_stop", TCL_GLOBAL_ONLY)))
	{
		quit = 1;
		return 1;
	}
	Tcl_DoOneEvent(TCL_DONT_WAIT);
	return 0;
}

void sendGui(const char *str)
{
	Tcl_SetVar(interp, "tr_bytes", str, TCL_GLOBAL_ONLY);
	Tcl_Eval(interp, "write_bytes");
}

